//! Here, all the embedded files are declared and initialized.

const std = @import("std");

/// Map that contains all available basis files.
pub const BASIS_FILES = std.StaticStringMap([]const u8).initComptime(.{
    .{"3-21g",       @embedFile("basis/3-21g.json"      )},
    .{"6-21g",       @embedFile("basis/6-21g.json"      )},
    .{"6-31g",       @embedFile("basis/6-31g.json"      )},
    .{"6-31g*",      @embedFile("basis/6-31gs.json"     )},
    .{"6-31g**",     @embedFile("basis/6-31gss.json"    )},
    .{"6-31+g",      @embedFile("basis/6-31pg.json"     )},
    .{"6-31+g*",     @embedFile("basis/6-31pgs.json"    )},
    .{"6-31+g**",    @embedFile("basis/6-31pgss.json"   )},
    .{"6-31++g",     @embedFile("basis/6-31ppg.json"    )},
    .{"6-31++g*",    @embedFile("basis/6-31ppgs.json"   )},
    .{"6-31++g**",   @embedFile("basis/6-31ppgss.json"  )},
    .{"6-311g",      @embedFile("basis/6-311g.json"     )},
    .{"6-311g*",     @embedFile("basis/6-311gs.json"    )},
    .{"6-311g**",    @embedFile("basis/6-311gss.json"   )},
    .{"6-311+g",     @embedFile("basis/6-311pg.json"    )},
    .{"6-311+g*",    @embedFile("basis/6-311pgs.json"   )},
    .{"6-311+g**",   @embedFile("basis/6-311pgss.json"  )},
    .{"6-311++g",    @embedFile("basis/6-311ppg.json"   )},
    .{"6-311++g*",   @embedFile("basis/6-311ppgs.json"  )},
    .{"6-311++g**",  @embedFile("basis/6-311ppgss.json" )},
    .{"aug-cc-pvdz", @embedFile("basis/aug-cc-pvdz.json")},
    .{"aug-cc-pvtz", @embedFile("basis/aug-cc-pvtz.json")},
    .{"aug-cc-pvqz", @embedFile("basis/aug-cc-pvqz.json")},
    .{"aug-cc-pv5z", @embedFile("basis/aug-cc-pv5z.json")},
    .{"cc-pvdz",     @embedFile("basis/cc-pvdz.json"    )},
    .{"cc-pvtz",     @embedFile("basis/cc-pvtz.json"    )},
    .{"cc-pvqz",     @embedFile("basis/cc-pvqz.json"    )},
    .{"cc-pv5z",     @embedFile("basis/cc-pv5z.json"    )},
    .{"def2-svp",    @embedFile("basis/def2-svp.json"   )},
    .{"def2-svpd",   @embedFile("basis/def2-svpd.json"  )},
    .{"def2-tzvp",   @embedFile("basis/def2-tzvp.json"  )},
    .{"def2-tzvpd",  @embedFile("basis/def2-tzvpd.json" )},
    .{"def2-tzvpp",  @embedFile("basis/def2-tzvpp.json" )},
    .{"def2-tzvppd", @embedFile("basis/def2-tzvppd.json")},
    .{"def2-qzvp",   @embedFile("basis/def2-qzvp.json"  )},
    .{"def2-qzvpd",  @embedFile("basis/def2-qzvpd.json" )},
    .{"def2-qzvpp",  @embedFile("basis/def2-qzvpp.json" )},
    .{"def2-qzvppd", @embedFile("basis/def2-qzvppd.json")},
    .{"sto-2g",      @embedFile("basis/sto-2g.json"     )},
    .{"sto-3g",      @embedFile("basis/sto-3g.json"     )},
    .{"sto-4g",      @embedFile("basis/sto-4g.json"     )},
    .{"sto-5g",      @embedFile("basis/sto-5g.json"     )},
    .{"sto-6g",      @embedFile("basis/sto-6g.json"     )}
});
